(* Generic database interface for mod_caml programs.
 * Copyright (C) 2003 Merjis Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: dbi.mli,v 1.2 2004/02/12 16:57:07 rwmj Exp $
 *)

(** Generic database interface.
  *
  * Making a connection to a specific type of database:
  *
  * {v
  * module DB = Dbi_postgres
  * let dbh = new DB.connection "database_name"
  * v}
  *
  * Equivalent to above, except that we make a connection to a named
  * type of database:
  *
  * {v
  * let dbh =
  *   try
  *     Dbi.Factory.connect "postgres" "database_name"
  *   with
  *     Not_found -> failwith "Postgres driver not available."
  * v}
  *
  * From Apache, using persistent connections (see [Apache.DbiPool]):
  *
  * {v
  * let dbh = Apache.DbiPool.get r "postgres" "database_name"
  * v}
  *
  * Simple usage, returning one row:
  *
  * {v
  * let sth = dbh#prepare "select name from employees where empid = ?" in
  * sth#execute [ `Int 100 ];
  * let row = sth#fetch1 in
  * let name = row.(0) in
  * v}
  *
  * Simple usage, returning multiple rows:
  *
  * {v
  * let sth = dbh#prepare "select name from employees where salary > ?" in
  * sth#execute [ `Int 10000 ];
  * sth#iter (fun row -> Printf.printf "name = %s\n" row.(0))
  * v}
  *
  * Advanced usage, binding columns for maximum efficiency:
  *
  * {v
  * let sth = dbh#prepare "select name from employees where salary > ?" in
  * sth#execute [ `Int 10000 ];
  * let name = ref "" in
  * sth#bind_columns [ `StringRef name ];
  * while sth#next do
  *   Printf.printf "name = %s\n" !name
  * done
  * v}
  *
  * Advanced usage, reusing prepared statements:
  *
  * {v
  * let sth =
  *   dbh#prepare "insert into employees (name, salary) values (?, ?)" in
  * List.iter (
  *   fun (name, salary) ->
  *     sth#execute [ `String name; `Int salary ];
  *     let id = sth#serial "" in
  *     Printf.printf "Employee %s has been assigned ID %d\n" name id
  * ) employee_list;
  * v}
  *)

type arg_t = [ `Null
	     | `Int of int
	     | `IntOption of int option
             | `Numeric of float
             | `NumericOption of float option
	     | `String of string
	     | `StringOption of string option
	     | `Timestamp of timestamp_t
	     | `TimestampOption of timestamp_t option
	     | `Interval of interval_t
	     | `IntervalOption of interval_t option
	     | `Bool of bool
	     | `BoolOption of bool option
	     ]
(** Type of arguments to the [statement#execute] method. The [*Option]
  * arguments can be [Some] thing or [None] meaning SQL [NULL].
  *)
and timestamp_t =
    {
      ts_is_null : bool;		(** Null? Other fields will be 0. *)
      ts_year : int;
      ts_month : int;
      ts_day : int;
      ts_hour : int;
      ts_min : int;
      ts_sec : int;
      ts_microsecs : int;
      ts_utc_offset : int;
    }
(** Timestamp type. *)
and interval_t =
    {
      iv_is_null : bool;		(** Null? Other fields will be 0. *)
      iv_years : int;
      iv_months : int;
      iv_days : int;
      iv_hours : int;
      iv_mins : int;
      iv_secs : int;
      iv_microsecs : int;
    }
(** Interval type. *)

val null_timestamp : timestamp_t
(** A timestamp with the [ts_is_null] field set to true. *)
val null_interval : interval_t
(** An interval with the [iv_is_null] field set to true. *)

type ref_t = [ `IntRef of int ref
             | `NumericRef of float ref
	     | `StringRef of string ref
		 (* XXX etc. *)
	     | `IntOptionRef of int option ref
             | `NumericOptionRef of float option ref
	     | `StringOptionRef of string option ref
		 (* XXX etc. *)
	     ]
(** Type of arguments to the [statement#bind_columns] method. *)

type precommit_handle
(** See {!Dbi.connection.register_precommit}. *)
type postrollback_handle
(** See {!Dbi.connection.register_postrollback}. *)

exception SQL_error of string
(** Exceptions thrown by subclasses on SQL errors. *)

class virtual statement :
  connection ->
object

  method virtual execute : arg_t list -> unit
    (** Execute the statement with the given list of arguments substituted
      * for [?] placeholders in the query string.
      *
      * This command can throw a variety of SQL-specific exceptions.
      *)

  method virtual fetch1 : string array
    (** The statement expects exactly one tuple to be returned from the query.
      * This returns the tuple, or throws [Not_found] if no tuple is
      * returned by the database.
      *)

  method fetchall : string array list
    (** This returns a list of all tuples returned from the query. Note that
      * this is less efficient than reading them one at a time, or using
      * [bind_columns] and [next].
      *)

  method iter : (string array -> unit) -> unit
    (** Iterate over the result tuples. *)

  method map : 'a . (string array -> 'a) -> 'a list
    (** Map over the result tuples. *)

  method fold_left : 'a . ('a -> string array -> 'a) -> 'a -> 'a
    (** Fold left over the result tuples. *)

  method fold_right : 'a . (string array -> 'a -> 'a) -> 'a -> 'a
    (** Fold right over the result tuples. NB: not tail recursive. *)

  method virtual bind_columns : ref_t list -> unit
    (** This method binds result columns to variables (references).
      * Subsequent calls to [next] will set those references for each
      * tuple returned. Setting references is much more efficient than
      * using the [fetch] methods, particularly when returning large
      * objects.
      *)

  method virtual next : bool
    (** Returns the [next] row. You would normally want to call
      * [bind_columns] first so that you can actually retrieve the
      * value of each column. [next] returns [true] if a rows was
      * returned or [false] otherwise.
      *)

  method virtual serial : string -> int
    (** If the statement is an INSERT and has been executed, then some
      * databases support retrieving the serial number of the INSERT
      * statement (assuming there is a SERIAL column or SEQUENCE attached
      * to the table). The string parameter is the sequence name, which
      * is only required by some database types. See the specific documentation
      * for your database for more information.
      *)

  method finish : unit
    (** "Finishes" the statement. This basically just frees up any memory
      * associated with the statement (this memory would be freed up by the
      * GC later anyway). After calling [finish] you may call [execute] to
      * begin executing another query.
      *)

  method connection : connection
    (** Return the database handle associated with this statement handle. *)
end

and virtual connection :
  ?host:string ->
  ?port:string ->
  ?user:string ->
  ?password:string ->
  string ->
object

  val mutable closed : bool

  method id : int
    (** Returns a unique integer which can be used to identify this
      * connection.
      *)

  method closed : bool
    (** Returns [true] if this database handle has been closed. Subsequent
      * operations on the handle will fail.
      *)

  method host : string option
    (** Return the [host] parameter. *)
  method port : string option
    (** Return the [port] parameter. *)
  method user : string option
    (** Return the [user] parameter. *)
  method password : string option
    (** Return the [password] parameter. *)
  method database : string
    (** Return the database name. *)

  method virtual database_type : string
    (** Database type (eg. "postgres"). *)

  method virtual prepare : string -> statement
    (** Prepare a database query, and return the prepared statement.
      * The statement may contain [?] placeholders which can be substituted
      * for values when the statement is executed.
      *)

  method prepare_cached : string -> statement
    (** This method is identical to [prepare] except that, if possible, it
      * caches the statement handle with the database object. Future calls
      * with the same query string return the previously prepared statement.
      * For databases which support prepared statement handles, this avoids
      * a round-trip to the database, and an expensive recompilation of the
      * statement.
      *)

  method ex : string -> arg_t list -> statement
    (** This is a shorthand for:
      *
      * {v
      * let sth = dbh#prepare_cached stmt in
      * sth#execute [args ...];
      * sth
      * v}
      *)

  method commit : unit
    (** Perform a COMMIT operation on the database. *)

  method rollback : unit
    (** Perform a ROLLBACK operation on the database. *)

  method register_precommit : (unit -> unit) -> precommit_handle
    (** Register a function which will be called just BEFORE a commit
      * happens on this handle. This method returns a handle which can
      * be used to deregister the callback later. This is useful for
      * implementing various types of persistence.
      *)

  method unregister_precommit : precommit_handle -> unit
    (** Unregister a precommit callback. *)

  method register_postrollback : (unit -> unit) -> postrollback_handle
    (** Register a function which will be called just AFTER a rollback
      * happens on this handle. This method returns a handle which can
      * be used to deregister the callback later. This is useful for
      * implementing various types of persistence.
      *)

  method unregister_postrollback : postrollback_handle -> unit
    (** Unregister a postrollback callback. *)

  method close : unit
    (** Closes the database handle. All statement handles are also invalidated.
      * Database handles which are collected by the GC are automatically
      * closed, but you should explicitly close handles to save resources,
      * where possible.
      *)

  method ping : bool
    (** This uses some active method to verify that the database handle is
      * still working. By this I mean that it tries to execute some sort of
      * 'null' statement against the database to see if it gets a response.
      * If the database is up, it returns [true]. If the database is down or
      * unresponsive, it returns [false]. This method should never throw
      * an exception (unless, perhaps, there is some sort of catastrophic
      * internal error in the [Dbi] library or the driver).
      *)

  method set_debug : bool -> unit
    (** Use this to enable debugging on the handle. In this mode significant
      * events (such as executing queries) are printed out on stderr.
      *)

  method debug : bool
    (** Returns true if this handle has debugging enabled. *)
end

module Factory :
  sig
    val connect : string ->
      ?host:string -> ?port:string -> ?user:string -> ?password:string ->
      string -> connection
      (** Connect to a specific type of database.  The first string parameter
	* is the database type, eg. "postgres", "mysql", etc.
	*
	* Throws [Invalid_argument] if the database type is not known.  May
	* throw other connection-specific SQL errors.
        *)

    val database_types : unit -> string list
      (** Returns a list of registered database types. *)

    val register : string -> (?host:string -> ?port:string ->
			       ?user:string -> ?password:string ->
			       string -> connection) -> unit
      (** Specific database drivers register themselves on load (or [Dynlink])
	* by calling this function.
        *)
  end
